"""
Utility functions for muldoon
"""

import numpy as np
from scipy.optimize import curve_fit
from scipy.stats import mode

def modified_lorentzian(t, baseline, slope, t0, DeltaP, Gamma):
    """
    Pressure profile for a vortex

    Args:
        t (float array): time
        baseline (float): pressure baseline against which vortex excursion
        occurs
        slope (float): slope against which excursion occurs
        t0 (float): central time for vortex excursion
        DeltaP (float): depth of pressure excursion
        Gamma (float): full-width/half-max duration of excursion

    Returns:
        Pressure excursion for a vortex (float array)

    """
    # Equation 7 from Kahapaa+ (2016)
    return baseline + slope*(t - t0) - DeltaP/(((t - t0)/(Gamma/2.))**2 + 1)

def redchisqg(ydata,ymod,deg=2,sd=None):
    """
    Returns the reduced chi-square error statistic for an arbitrary model,
    chisq/nu, where nu is the number of degrees of freedom. If individual
    standard deviations (array sd) are supplied, then the chi-square error
    statistic is computed as the sum of squared errors divided by the standard
    deviations. See http://en.wikipedia.org/wiki/Goodness_of_fit for reference.

    ydata,ymod,sd assumed to be Numpy arrays. deg integer.

    Usage:
    chisq=redchisqg(ydata,ymod,n,sd)

    where
    ydata : data
    ymod : model evaluated at the same x points as ydata
    n : number of free parameters in the model
    sd : uncertainties in ydata

    Rodrigo Nemmen
    http://goo.gl/8S1Oo
    """
    # Chi-square statistic
    if(np.any(sd == None)):
        chisq=np.sum((ydata-ymod)**2)
    else:
        chisq=np.sum( ((ydata-ymod)/sd)**2 )

    # Number of degrees of freedom assuming 2 free parameters
    nu=ydata.size - 1. - deg

    return chisq/nu


def fit_vortex(vortex, init_params, bounds, sigma=None, 
        rescale_uncertainties=True):
    """
    Fits modified Lorentzian to pressure profile

    Args:
        vortex (dict of float arrays): ["time"] - times, ["pressure"] - pressures
        init_params (float array): initial values including baseline, slope, initial central time, initial delta P, and initial duration
        bounds (float array): bounds on fit parameters listed in the same order as in init_params
        sigma (float, float array, optional): per-point uncertainties
        rescale_uncertainties (bool, optional): whether to rescale uncertainties on fit parameters by sqrt(reduced chi-squared)

    Returns:
        fit parameters (float array) and uncertainties (float array)

    """

    x = vortex["time"]
    y = vortex["pressure"]

    if(sigma is not None):
        popt, pcov = curve_fit(modified_lorentzian, x, y, p0=init_params,
                bounds=bounds, sigma=sigma)
    else:
        popt, pcov = curve_fit(modified_lorentzian, x, y, p0=init_params,
                bounds=bounds)
    ymod = modified_lorentzian(x, *popt)

    if(rescale_uncertainties):
        if(sigma is None):
            sd = np.nanstd(y - ymod)
        else:
            sd = sigma
        red_chisq = redchisqg(y, ymod, deg=5, sd=sd)

        pcov *= np.sqrt(red_chisq)

    return popt, np.sqrt(np.diag(pcov))

def write_out_plot_data(x, y, x_label, y_label, 
        xerr=None, yerr=None, filename="out.csv", mode="w", test_mode=False):
    """
    Write out formatted text file of plot data

    Args:
        x/y (float array): x/y points to write out
        x_label/y_label (str): labels for columns
        xerr/yerr (float array, optional): associated uncertainties
        filename (str, optional): path of file to which to write out data
        mode (str, optional): write mode; defaults to over-write
        test_mode (bool, optional): whether to actually write out file

    """

    # Construct write string
    write_str = "# %s, %s" % (x_label, y_label)

    if(xerr is not None):
        write_str += ", %s_err" % x_label
    if(yerr is not None):
        write_str += ", %s_err" % y_label

    write_str += "\n"

    for i in range(len(x) - 1):
        write_str += "%g, %g" % (x[i], y[i])

        if(xerr is not None):
            write_str += ", %g" % (xerr[i])
        if(yerr is not None):
            write_str += ", %g" % (yerr[i])

        write_str += "\n"

    # Don't write a new-line character for the last entry
    write_str += "%g, %g" % (x[-1], y[-1])
    if(xerr is not None):
        write_str += ", %g" % (xerr[i])
    if(yerr is not None):
        write_str += ", %g" % (yerr[i])

    if(~test_mode):
        f = open(filename, mode)
        f.write(write_str)
        f.close()

    return write_str

def find_gaps(time):
    """
    Finds gaps in the time-series

    Args:
        time (float array): time array

    """

    # Calculate all the delta t's
    delta_ts = (time[1:] - time[0:-1])
    ind = delta_ts > 0.

    # Are there any delta t's bigger than the typical sampling?
    mod = mode(delta_ts[ind])[0][0]

    # If there are no gaps in the time-series
    ind = np.argwhere(~np.isclose(delta_ts, mod, rtol=0., atol=mod))[:,0]
    if(len(delta_ts[ind]) == 0):
        return []
    else:
        return ind

def break_at_gaps(time, data):
    """
    Break the time-series into pieces if there are gaps

    Args:
        time/data (float arrays): time and data time-series

    Returns:
        the time-series broken into lists at the gaps

    """

    gaps = find_gaps(time)

    if(len(gaps) == 0):
        return [time], [data]
    else:
        ret_time = list()
        ret_data = list()

        last_gap = 0
        for i in range(len(gaps)):

            ret_time.append(time[last_gap:gaps[i]+1])
            ret_data.append(data[last_gap:gaps[i]+1])

            last_gap = gaps[i]+1

        ret_time.append(time[gaps[-1]+1:])
        ret_data.append(data[gaps[-1]+1:])

        return ret_time, ret_data

def plot_vortex(time, t0, data, model_func, popt, ax, yerr=None):
    """
    Make plot of vortex and model

    Args:
        time/data (float arrays): time-series to plot and model fit
        t0 (float): the central time of the vortex signal
        model_func (function): function that will generate model
        popt (float array): model parameters
        ax (matplotlib axis): ax to which to plot
    
    Returns:
        model (float array)

    """

    # Boise State official colors in hex
    # boisestate.edu/communicationsandmarketing/brand-standards/colors/
    BoiseState_blue = "#0033A0"
    BoiseState_orange = "#D64309"

    ### Fit vortex ###
    model = model_func(time, popt)
    if(yerr is not None):
        ax.errorbar((time - t0)*3600., data, yerr=yerr, 
                ls='', marker='o', color=BoiseState_blue)
    else:
        ax.plot((time - t0)*3600., data, 
                ls='', marker='o', color=BoiseState_blue)

    # Plot model fit
    ax.plot((time - t0)*3600., model, lw=3, color=BoiseState_orange, zorder=-1)

    return model



"""A package for sifting meteorological data to look for and analyze vortex
encounters"""

__version__ = '0.5.2'
__author__ = 'Brian Jackson <bjackson@boisestate.edu>'
__all__ = ['met_timeseries', 'utils', 'read_data']

from .met_timeseries import *
from .utils import *
from .read_data import *



"""
A collections of routines to read in data from various missions
"""

import numpy as np
import pandas as pd

def read_Perseverance_MEDA_data(filename, sol=None):
    """
    Read in Perseverance MEDA data - https://pds-atmospheres.nmsu.edu/PDS/data/PDS4/Mars2020/mars2020_meda/

    Args:
        filename (str): path to CSV file

    Returns:
        time, pressure (float array): times and pressures, times in seconds
        since midnight of sol associated with filename
    """

    time = make_seconds_since_midnight(filename)
    pressure = pd.read_csv(filename)['PRESSURE'].values

    return time, pressure

def make_seconds_since_midnight(filename):
    """
    The MEDA data provide times in the LTST field in the format "sol hour:minute:second".

    Args:
        filename (str): name of the file

    Returns:
        number of seconds in each row since midnight of the primary sol for that file
        """

    if(sol is None):
        primary_sol = which_sol(filename)
    data = pd.read_csv(filename)

    # Grab the sols and times associated with each row
    sols_str = data['LTST'].str.split(expand=True)[0].values
    times_str = data['LTST'].str.split(expand=True)[1].values

    # Turn the times strings into seconds since midnight of the primary sol
    delta_sols = sols_str.astype(float) - float(primary_sol)

    # And then in a very cludgey way, convert times to seconds since primary sol's midnight
    time = np.array([float(times_str[i].split(":")[0]) + delta_sols[i]*24.  +\
            float(times_str[i].split(":")[1])/60 +\
            float(times_str[i].split(":")[2])/3600. for i in
            range(len(times_str))])

    return time

def which_sol(filename):
    """
    Based on the filename, returns the sol corresponding to a data file

    Args:
        filename (str): name of the file

    Returns:
        sol (int) associated with that file
    """

    ind = filename.find("WE__")

    return int(filename[ind+4:ind+8])



"""
Test the functionality of Muldoon
"""

import numpy
import numpy as np
from numpy.random import normal
from muldoon import met_timeseries as met
from muldoon import utils 

# Create time-series
time = np.linspace(-1, 1, 1000)
baseline = 0.
slope = 1.
t0 = 0.
DeltaP = 1.
Gamma = 0.01
right_answer = np.array([baseline, slope, t0, DeltaP, Gamma])
profile = utils.modified_lorentzian(time, baseline, slope, t0, DeltaP, Gamma) +\
    normal(scale=slope/20., size=len(time))
mt = met.MetTimeseries(time, profile)

# Detrend
window_size = Gamma
detrended_pressure = mt.detrend_pressure_timeseries(window_size)

# Calculate filter
conv = mt.apply_lorentzian_matched_filter(2.*mt.sampling, 1./np.pi)

# Max mis-match between fit and right answers allowed
num_sigma = 5.

def test_detrend_pressure_timeseries():

    # Make sure detrend is within 
    assert np.isclose(np.std(mt.detrended_pressure), 0.1, atol=0.1)

def test_apply_lorentzian_matched_filter():
    # Test the matched filter analysis

    # Find the maximum
    mx_ind = np.argmax(mt.convolution)

    # Make sure convolution returns a strong peak at the right time
    assert ((np.abs(mt.time[mx_ind]) < 2.*Gamma) &\
            (mt.convolution[mx_ind] > 5.))

def test_find_vortices():

    # Test find_vortices
    vortices = mt.find_vortices()
    # Make sure the max peak in the convolution is where the vortex is
    assert(mt.time[mt.peak_indices[0]] < 2.*Gamma)

def test_fit_vortex():

    vortices = mt.find_vortices()

    # Test vortex fit
    old_popt, old_unc = utils.fit_vortex(vortices[0], [0., 1., 0., 1., 0.01], 
                          [[-1, -1, np.min(vortices[0]["time"]), 0, 0],
                           [1, 1, np.max(vortices[0]["time"]), 2, 1]],
                          sigma=vortices[0]["pressure_scatter"])

    # Make sure best-fit parameters all match the right answers 
    assert(np.max(np.abs(old_popt - right_answer)/old_unc) < num_sigma)

def test_init_params_bounds():
    vortices = mt.find_vortices()

    # Test fit parameter initialization and bounds calculation
    init_params = mt._determine_init_params(vortices[0])
    bounds = mt._determine_bounds(vortices[0], init_params)
    
    popt, unc = utils.fit_vortex(vortices[0], init_params, bounds)

    # Make sure best-fit parameters all match the right answers
    assert(np.max(np.abs(popt - right_answer)/unc) < num_sigma)

def test_fit_all_vortices():
    # Make sure fit_all_vortices works, too
    vortices = mt.find_vortices()
    popts, uncs = mt.fit_all_vortices()

    # Make sure best-fit parameters all match the right answers
    assert(np.max(np.abs(popts[0] - right_answer)/uncs[0]) < num_sigma)



import numpy as np
from numpy.random import choice
import matplotlib.pyplot as plt
from scipy.signal import find_peaks, peak_widths, boxcar
from astropy.convolution import convolve as astropy_convolve
from scipy.stats import mode
from statsmodels.robust import mad

import muldoon.utils as utils

__all__ = ['MetTimeseries']


class MetTimeseries(object):
    """
    Process and analyze a meteorological time-series to search for vortices
    """

    def __init__(self, time, pressure, windspeed=None, wind_direction=None):
        """
        Args:
            time (float, array): time of meteorological time-series
            pressure (float, array): pressure measurements
            windspeed (float, array, optional): wind speed measurements
            wind_direction (float, array, optional): wind velocity aziumth
        """

        self.time = time
        # Calculate the sampling rate
        self.sampling = mode(time[1:] - time[0:-1]).mode[0]

        self.pressure = pressure

        if(windspeed is not None):
            self.windspeed = windspeed

        if(wind_direction is not None):
            self.wind_direction = wind_direction

        # Filtered pressure time-series
        self.detrended_pressure = None
        self.window_width = None
        self.detrended_pressure_scatter = None

        # Time-series filters
        self.pressure_trend = None

        # matched filter parameters
        self.matched_filter_fwhm = None
        self.matched_filter_depth = None

        # detection threshold
        self.detection_threshold = None

        # FWHMs for vortices that are detected
        self.num_fwhms = None

        # convolution of matched filter
        self.convolution = None

        self.peak_indices = None
        self.peak_widths = None

        # Collection of time-series for individual vortices
        self.vortices = None

        # Vortex fit parameters and uncertainties
        self.popts = None
        self.uncs = None

    def detrend_pressure_timeseries(self, window_width, 
            deal_with_gaps=True):
        """
        Applies boxcar filter to pressure time-series

        Args:
            window_width (float): width of window in the same units as time
            deal_with_gaps (bool, optional): whether to check for gaps

        Returns:
            detrended pressure time-series (float array)

        """

        self.window_width = window_width

        # Queue up to deal with gaps
        local_time = [self.time]
        local_pressure = [self.pressure]

        if(deal_with_gaps):
            local_time, local_pressure = utils.break_at_gaps(self.time,
                    self.pressure)
            
        # Calculate number of points for window_width
        window_size = int(window_width/self.sampling)
        # Check that window_size is odd
        if(window_size % 2 == 0): 
            window_size += 1

        self.window_size = window_size

        # Empty out the Nones
        self.detrended_pressure = np.array([])

        # Detrend, piece at a time
        for i in range(len(local_time)):

            local_pressure_trend = astropy_convolve(local_pressure[i],
                    boxcar(window_size), boundary='extend', 
                    preserve_nan=True)

            if(self.pressure_trend is None):
                self.pressure_trend = local_pressure_trend
            else:
                self.pressure_trend = np.append(self.pressure_trend,
                        local_pressure_trend)

            self.detrended_pressure =\
                    np.append(self.detrended_pressure, 
                            local_pressure[i] - local_pressure_trend)

        self.detrended_pressure_scatter = np.nanstd(self.detrended_pressure)

        return self.detrended_pressure

    def write_out_detrended_timeseries(self, filename="out.csv", mode="w",
            test_mode=False):
        """
        Write out formatted text file of detrended time-series

        Args:
            filename (str, optional): path of file to which to write out data
            mode (str, optional): write mode; defaults to over-write
            test_mode (bool, optional): whether to actually write out file

        """

        if(self.detrended_pressure is None):
            raise ValueError("Need to detrend pressure!")

        # Construct write string
        write_str = "# time, pressure\n"

        for i in range(len(self.time) - 1):
            write_str += "%g, %g\n" %\
                    (self.time[i], self.detrended_pressure[i])

        # Don't write a new-line character for the last entry
        write_str += "%g, %g" % (self.time[-1], self.detrended_pressure[-1])
            
        if(~test_mode):
            f = open(filename, mode)
            f.write(write_str)
            f.close()

        return write_str

    def apply_lorentzian_matched_filter(self, lorentzian_fwhm=None,
            lorentzian_depth=1./np.pi, num_fwhms=6.):
        """
        Applies Lorentzian matched filter to detrended pressure to find
        vortices

        Args:
            lorentzian_fwhm (float, optional): the full-width/half-max of the matched filter
            lorentzian_depth (float, optional): the depth of the filter; probably wants to be 1./np.pi
            num_fwhms (float, optional): how many full-width/half-maxes to
            generate matched filter; defaults to 6

        Returns:
            Results from matched filter (float array)

        """

        if(self.detrended_pressure_scatter is None):
            raise ValueError("Run detrend_pressure_timeseries first!")

        if(lorentzian_fwhm is None):
            lorentzian_fwhm = 2.*self.sampling

        self.matched_filter_fwhm = lorentzian_fwhm
        self.matched_filter_depth = lorentzian_depth

        self.num_fwhms = num_fwhms

        lorentzian_time = np.arange(-num_fwhms/2*lorentzian_fwhm, 
                num_fwhms/2.*lorentzian_fwhm, self.sampling)
        lorentzian = utils.modified_lorentzian(lorentzian_time, 0., 0., 0., 
                lorentzian_depth, lorentzian_fwhm)

        # Make sure the matched filter isn't wider than the signal itself
        if(len(lorentzian_time) > len(self.time)):
            raise ValueError("lorentzian_time is wider than detrended "+\
                    "pressure!")

        convolution =\
            np.convolve(self.detrended_pressure/\
            self.detrended_pressure_scatter, 
                    lorentzian, mode='same')

        # Shift and normalize
        med = np.nanmedian(convolution)
        md = mad(convolution)
        self.convolution = (convolution - med)/md

        return self.convolution

    def find_vortices(self, detection_threshold=5, distance=20, fwhm_factor=6):
        """
        Finds distinct peaks in the matched-filter convolution, presumably
        vortex signals

        Args: 
            detection_threshold (float, optional): threshold for peak detection
            distance (int, optional): min number of point between peaks
            fwhm_factor (int, optional): when returning vortices, how wide a time window to return

        Returns:
            list of times and pressures for each vortex

        """

        self.detection_threshold = detection_threshold

        if(self.convolution is None):
            raise ValueError("Run apply_lorentzian_matched_filter first!")

        ex = find_peaks(self.convolution, distance=distance)
        ind = self.convolution[ex[0]] >= detection_threshold

        pk_wds, _, _, _ = peak_widths(self.convolution, ex[0][ind])

        # Sort from largest peak to smallest
        srt_ind = np.argsort(self.convolution[ex[0][ind]])[::-1]
        self.peak_indices = ex[0][ind][srt_ind]
        self.peak_widths = pk_wds[srt_ind]

        self.vortices = []
        for i in range(len(self.peak_indices)):
            mn_ind = int(self.peak_indices[i] -\
                    fwhm_factor/2*int(self.peak_widths[i]))
            mx_ind = int(self.peak_indices[i] +\
                    fwhm_factor/2*int(self.peak_widths[i]))

            # Make sure there are enough points in the vortex
            if(mx_ind - mn_ind > 5):
                # Use original, unfiltered data
                self.vortices.append({"time": self.time[mn_ind:mx_ind],
                    "pressure": self.pressure[mn_ind:mx_ind],
                    "pressure_scatter": self.detrended_pressure_scatter*\
                            np.ones_like(self.time[mn_ind:mx_ind])})

        return self.vortices

    def fit_all_vortices(self, 
            filepath=None, figsize=(10, 10), aspect_ratio=16./9,
            pressure_units="Pa", time_units="Hours", vortex_time_units="s"):
        """
        Fit all vortices with modified Lorentzian and return fit parameters and
        uncertainties

        Args:
            filepath (str, optional): If not None, gives the filepath and filename stem to which to write vortex figures
            fig (matplotlib figure, optional): figure into which to plot
            figsize (float tuple, optional): size of figure
            aspect_ratio (float, optional): figure aspect ratio
            pressure/time/vortex_time_units (str, optional): axes labels

        Returns:
            list of two arrays, the first with best fit parameters and the
            second with uncertainties

        """

        if(self.vortices is None):
            raise ValueError("Run find_vortices first!")
        if(len(self.vortices) == 0):
            raise ValueError("There are no vortices!")

        self.popts = list()
        self.uncs = list()
        for i in range(len(self.vortices)):
            # Estimate initial parameters
            init_params = self._determine_init_params(self.vortices[i])

            # Estimate bounds
            bounds = self._determine_bounds(self.vortices[i], init_params)

            try:
                popt, unc = utils.fit_vortex(self.vortices[i], init_params, 
                        bounds, sigma=self.vortices[i]["pressure_scatter"])

                self.popts.append(popt)
                self.uncs.append(unc)

                # Make figure
                if(filepath is not None):
                    fig = plt.figure(figsize=(figsize[0]*aspect_ratio,
                        figsize[1]))

                    ax = fig.add_subplot(111)

                    # Remember! Times are in hours!
                    time = self.vortices[i]["time"]
                    ydata = self.vortices[i]["pressure"] - popt[0]
                    yerr = self.vortices[i]["pressure_scatter"]
                    model_func = lambda time, popt:\
                            utils.modified_lorentzian(time, *popt) - popt[0]
                    model = utils.plot_vortex(time, 
                            popt[2], ydata, model_func, popt, ax, yerr=yerr)

                    ax.text(0.05, 0.05, "%s, vortex %i" % (filepath, i),
                            fontsize=24, transform=ax.transAxes)
                    ax.grid(True)
                    ax.tick_params(labelsize=24)
                    ax.set_xlabel(r'$t - t_0\,\left( {\rm %s} \right)$' %\
                        (vortex_time_units), fontsize=36)
                    ax.set_ylabel(r'$\Delta P\,\left( {\rm %s} \right)$' %\
                        (pressure_units), fontsize=36)

                    fig.savefig("%s_vortex%i.jpg" % (filepath, i), 
                            bbox_inches="tight", dpi=500)

                    # Write out data
                    filename = filepath + "_data_vortex%i.csv" % (i)
                    utils.write_out_plot_data((time - popt[2])*3600., 
                            ydata - popt[0],
                            "Time", "DeltaP", yerr=yerr, filename=filename)

                    filename = filepath + "_model_vortex%i.csv" % (i)
                    utils.write_out_plot_data((time - popt[2])*3600., model, 
                            "Time", "DeltaP", filename=filename)

            except:
                print("fit_all_vortices: vortex %i couldn't be fit!" % i)

            # Reset ax
            fig = None
            ax = None
        return self.popts, self.uncs

    def _determine_init_params(self, vortex, 
            init_baseline=None, init_slope=None, init_t0=None, 
            init_DeltaP=None, init_Gamma=None):
        """
        Estimate reasonable initial parameters for fitting a vortex pressure
        signal

        Args:
            vortex (dict of float arrays): vortex["time"] - time, 
            vortex["pressure"] - pressure
            init_* (float): initial parameters

        Returns:
            float array of initial parameter values

        """

        x = vortex["time"]
        y = vortex["pressure"]

        # Initial fit to background trend
        fit_params = np.polyfit(x, y, 1)
        detrended_y = y - np.polyval(fit_params, x)

        if(init_baseline is None):
            init_baseline = np.median(y)

        if(init_slope is None):
            init_slope = (y[-1] - y[0])/(x[-1] - x[0])

        if(init_t0 is None):
            init_t0 = x[np.argmin(detrended_y)]

        if(init_DeltaP is None):
            init_DeltaP = np.max(detrended_y) - np.min(detrended_y) 

        if(init_Gamma is None):
            init_Gamma = 5.*self.sampling

        return np.array([init_baseline, init_slope, init_t0, init_DeltaP, 
            init_Gamma])

    def _determine_bounds(self, vortex, init_params,
            slope_fac=10., Gamma_fac=10.):
        """
        Estimate reasonable bounds on fit parameters

        Args:
            vortex (dict of float arrays): vortex["time"] - time,
            vortex["pressure"] - pressure
            init_params (float array): initial parameters in following order:
                init_baseline, init_slope, init_t0, init_DeltaP, init_Gamma
            slope_fac (float): maximum factor for slope upper bound
            Gamma_fac (float): maximum factor for Gamma upper bound
    
        Returns:
            float array with lower and upper bounds on fit parameters

        """

        x = vortex["time"]
        y = vortex["pressure"]

       # Initial fit to background trend
        fit_params = np.polyfit(x, y, 1)
        detrended_y = y - np.polyval(fit_params, x)

        # Baseline probably doesn't exceed minimum or maximum y
        mn_baseline = np.min(y)
        mx_baseline = np.max(y)

        # Slope unlikely to exceed overall slope
        overall_slope = (y[-1] - y[0])/(x[-1] - x[0])
        mn_slope = -slope_fac*np.abs(overall_slope)
        mx_slope = slope_fac*np.abs(overall_slope)

        mn_t0 = np.min(x)
        mx_t0 = np.max(x)

        # Can't have negative delta P's
        mn_deltaP = 0.
        mx_deltaP = np.max(detrended_y) - np.min(detrended_y)

        mn_Gamma = 2.*self.sampling # Nyquist sampling
        mx_Gamma = np.max([Gamma_fac*init_params[4], x[-1] - x[0]])

        return ([mn_baseline, mn_slope, mn_t0, mn_deltaP, mn_Gamma],
                [mx_baseline, mx_slope, mx_t0, mx_deltaP, mx_Gamma])

    def make_conditioned_data_figure(self, which_vortex=0, 
            fig=None, figsize=(10, 10), aspect_ratio=16./9,
            pressure_units="Pa", time_units="Hours", vortex_time_units="s",
            write_filename=None):
        """
        Make figure showing the data conditioning and analysis process -
        like Figure 1 of Jackson et al. (2021)

        Args:
            which_vortex (int): which vortex within vortices to plot
            fig (matplotlib figure obj, optional): the figure object to use
            figsize (2x1 list, optional): inches x inches figure size
            aspect_ratio (float, optional): figure aspect ratio
            pressure/time/vortex_time_units (str, optional): units to label axes
            write_filename (str, optional): filename stem to use for writing out

        Returns:
            figure and all axes

        """

        # Boise State official colors in hex
        # boisestate.edu/communicationsandmarketing/brand-standards/colors/
        BoiseState_blue = "#0033A0"
        BoiseState_orange = "#D64309"

        if(self.peak_indices is None):
            raise ValueError("Run find_vortices first!")
        if(self.popts is None):
            raise ValueError("Run fit_all_vortices first!")

        if(fig is None):
            fig = plt.figure(figsize=(figsize[0]*aspect_ratio, figsize[1]))

        # Add axes
        ax1 = fig.add_subplot(221)
        ax2 = fig.add_subplot(223, sharex=ax1)
        ax3 = fig.add_subplot(222)
        ax4 = fig.add_subplot(224)

        ### Raw data ###
        ax1.plot(self.time, self.pressure, 
                marker='.', ls='', color=BoiseState_blue)
        ax1.text(0.05, 0.8, "(a)", fontsize=48, transform=ax1.transAxes)
        ax1.grid(True)
        ax1.tick_params(labelsize=24, labelbottom=False)

        if(write_filename is not None):
            filename = write_filename + "panel_a.csv"
            utils.write_out_plot_data(self.time, self.pressure, 
                    "Time", "Pressure", filename=filename)

        ### Filtered data ###
        ax2.plot(self.time, self.detrended_pressure, 
                marker='.', ls='', color=BoiseState_blue)
        ax2.text(0.05, 0.05, "(b)", fontsize=48, transform=ax2.transAxes)
        ax2.grid(True)
        ax2.tick_params(labelsize=24)
        ax2.set_xlabel("Time (%s)" % (time_units), fontsize=36)
        ax2.set_ylabel(r'$\Delta P\,\left( {\rm %s} \right)$' %\
                (pressure_units), fontsize=36)

        if(write_filename is not None):
            filename = write_filename + "panel_b.csv"
            utils.write_out_plot_data(self.time, self.detrended_pressure,
                    "Time", "Detrended_Pressure", filename=filename)


        ### Convolution ###
        ax3.plot(self.time, self.convolution, 
                color=BoiseState_blue, ls='', marker='.')
        ax3.text(0.05, 0.8, "(c)", fontsize=48, transform=ax3.transAxes)
        ax3.grid(True)
        ax3.yaxis.set_label_position("right")
        ax3.yaxis.tick_right()
        ax3.tick_params(labelsize=24, labelleft=False, labelright=True)
        ax3.set_ylabel(r'$\left( F \ast \Delta P \right)$', fontsize=36)

        if(write_filename is not None):
            filename = write_filename + "panel_c.csv"
            utils.write_out_plot_data(self.time, self.convolution,
                    "Time", "Convolution", filename=filename)


        # Add lines in all plots highlighting the detections
        for cur_ex in self.peak_indices:
            ax1.axvline(self.time[cur_ex], 
                    color=BoiseState_orange, zorder=-1, ls='--', lw=3)
            ax2.axvline(self.time[cur_ex], 
                    color=BoiseState_orange, zorder=-1, ls='--', lw=3)
            ax3.axvline(self.time[cur_ex], 
                    color=BoiseState_orange, zorder=-1, ls='--', lw=3)


        ### Fit vortex ###
        model_func = lambda time, popt:\
                utils.modified_lorentzian(time, *popt) - popt[0]

        # Remember! Times are in hours!
        time = self.vortices[which_vortex]["time"]
        ydata = self.vortices[which_vortex]["pressure"] -\
                self.popts[which_vortex][0]
        yerr = self.vortices[which_vortex]["pressure_scatter"]
        model = utils.plot_vortex(time, self.popts[which_vortex][2], ydata, 
                model_func, self.popts[which_vortex], ax4, yerr=yerr)

        ax4.text(0.05, 0.05, "(d)", fontsize=48, transform=ax4.transAxes)
        ax4.grid(True)
        ax4.yaxis.set_label_position("right")
        ax4.yaxis.tick_right()
        ax4.tick_params(labelsize=24, labelleft=False, labelright=True)

        ax4.set_xlabel(r'$t - t_0\,\left( {\rm %s} \right)$' %\
                (vortex_time_units), fontsize=36)
        ax4.set_ylabel(r'$\Delta P\,\left( {\rm %s} \right)$' %\
                (pressure_units), fontsize=36)

        if(write_filename is not None):
            filename = write_filename + "panel_d_data.csv"
            utils.write_out_plot_data(x, ydata, "Time", "DeltaP", 
                    yerr=yerr, filename=filename)

            filename = write_filename + "panel_d_model.csv"
            utils.write_out_plot_data(x, vortex_model, "Time", "DeltaP", 
                    filename=filename)

        return fig, ax1, ax2, ax3, ax4

    def pressure_timeseries_injection_recovery(self,
        mn_deltaP=0.1, mx_deltaP=10., num_deltaPs=10,
        mn_Gamma=2.0, mx_Gamma=200., num_Gammas=10,
        num_trials=10,
        num_fwhms=3.):
        """
        Performs an injection/recovery for vortex signal within a given raw pressure time-series

        Args:
            mn/mx (floats, optional): min/max values for deltaP and Gamma arrays
            num (int, optional): number of points in the deltaP/Gamma arrays
            num_trials (int, optional): number of trials at each deltaP/Gamma
            num_fwhms (float, optional): number of FWHMs for matched_filter

        Returns:
            deltaP, Gamma, and detection fraction arrays (float arrays)
        """

        deltaP_grid = 10.**(np.linspace(np.log10(mn_deltaP), np.log10(mx_deltaP), num_deltaPs))
        Gamma_grid = 10.**(np.linspace(np.log10(mn_Gamma), np.log10(mx_Gamma), num_Gammas))/3600. # convert to hours

        # Whether or not a detection was made
        detection_statistics =\
                np.zeros([num_deltaPs, num_Gammas, num_trials])
        for i in range(num_deltaPs):
            for j in range(num_Gammas):
                for k in range(num_trials):

                    # Generate vortex at random time
                    correct_t0 = choice(self.time)
                    vortex_signal = utils.modified_lorentzian(self.time, 0., 
                            0., correct_t0, deltaP_grid[i], Gamma_grid[j])

                    # Inject into flipped detrended time-series
                    synthetic_time_series = self.detrended_pressure +\
                            vortex_signal + self.pressure_trend

                    # Make new object
                    new_mt = MetTimeseries(self.time, synthetic_time_series)
                    new_mt.detrend_pressure_timeseries(self.window_width)

                    new_mt.apply_lorentzian_matched_filter(
                            self.matched_filter_fwhm, 
                            self.matched_filter_depth, num_fwhms=self.num_fwhms)

                    new_mt.find_vortices(detection_threshold=\
                            self.detection_threshold)

                    # If find_vortices found the vortex
                    if(np.any(np.abs(new_mt.time[new_mt.peak_indices] -\
                            correct_t0) < Gamma_grid[j])):
                        detection_statistics[i,j,k] = 1.
        detection_statistics = np.mean(detection_statistics, axis=2)

        # Convert Gamma back to seconds
        return deltaP_grid, Gamma_grid*3600., detection_statistics
